/*
   $Header: /u/drspeech/repos/rtst/rtst_test.c,v 1.23 1996/10/15 23:49:10 davidj Exp $

   Some basic tests of "rtst" functionality.
*/

#include <stdio.h>
#include "rtst.h"

/* For systems with broken header files. */
extern void srand48(long);
extern long lrand48(void);
extern unsigned short *seed48(unsigned short seed16v[3]);
extern void *malloc(size_t);
extern void free(void *);

#define PADVEC_MAX_ALLOCS (64)

void
padvec_test(void)
{
    int i;
    void *p[PADVEC_MAX_ALLOCS];

    rtst_start("padvec_test");
    rtst_log("Testing alloc/write/dealloc of padvecs.\n");

    for (i=0; i<rtst_numtests; i++)
    {
        unsigned long n_allocs = (unsigned long) lrand48() % PADVEC_MAX_ALLOCS;
        unsigned long j;

        rtst_log("Starting round i=%d, n_allocs=%ld \n", i, n_allocs);

        /* Allocate set of padvecs. */
        for (j=0; j<n_allocs; j++)
        {
            const int align = 1 << (lrand48() % 6);
            const size_t len = (lrand48() % 1000) + 1;
            
            p[j] = rtst_padvec_new(len, align);

            /* Scribble in the middle somewhere. */
            ((unsigned char*) p[j])[lrand48() % len] = lrand48();
        }
        
        /* Then dellocate */
        for (j=0; j<n_allocs; j++)
        {
            rtst_padvec_del(p[j]);
        }
    }
    rtst_passed();
}

/* Test clone functions. */

void
padvec_clone_test(void)
{
    int i;

    rtst_start("padvec_clone_test");
    rtst_log("Testing padvec clone operations.\n");

    for (i=0; i<rtst_numtests; i++)
    {
        int align;
        size_t len;
        void *orig_p;
        void *copy_p;

	align = 1 << (lrand48() % 6);
	len = (lrand48() % 1000) + 1;
        orig_p = malloc(len);
        rtst_urand_i8i8_vi8(len, -0x80, 0x7f, (rtst_int8 *) orig_p);

        copy_p = rtst_padvec_clone(len, align, orig_p);

        rtst_checkeq_vi8vi8(len, orig_p, copy_p);

        free(orig_p);
        rtst_padvec_del(copy_p);
    }
    rtst_passed();
}

void
padmat_test(void)
{
    int i;
    void *p[PADVEC_MAX_ALLOCS];

    rtst_start("padmat_test");
    rtst_log("Testing alloc/write/dealloc of padmats.\n");

    for (i=0; i<rtst_numtests; i++)
    {
        unsigned long n_allocs = (unsigned long) lrand48() % PADVEC_MAX_ALLOCS;
        unsigned long j;

        rtst_log("Starting round i=%d, n_allocs=%ld \n", i, n_allocs);

        /* Allocate set of padmats. */
        for (j=0; j<n_allocs; j++)
        {
            int align;
            size_t rows;
            size_t cols;
            size_t margin;
	    size_t lda;
	    size_t offset;

	    align = 1 << (lrand48() % 6);
            rows = (lrand48() % 100) + 1;
            cols = (lrand48() % 100) + 1;
            margin = (lrand48() % 100) + 1;
	    lda = cols + margin;
/*	    
	    rtst_log("Allocating padmat, rows=%lu cols=%lu lda=%lu align=%d\n",
		     (unsigned long) rows, (unsigned long) cols,
		     (unsigned long) lda, align);
*/
            p[j] = rtst_padmat_new(rows, cols, lda, align);

            /* Scribble in the middle somewhere. */
	    offset = (lrand48() % rows) * lda + (lrand48() % cols);
            ((unsigned char*) p[j])[offset] = (unsigned char) lrand48(); 
        }
        
        /* Then dellocate */
        for (j=0; j<n_allocs; j++)
        {
            rtst_padmat_del(p[j]);
        }
    }
    rtst_passed();
}

/* Test clone functions. */

void
padmat_clone_test(void)
{
    int i;
    size_t j;
    int align;
    size_t rows;
    size_t cols;
    size_t margin;
    size_t lda;
    size_t len;
    void *orig_p;
    void *copy_p;
    rtst_int8 *orow_p;
    rtst_int8 *crow_p;

    rtst_start("padmat_clone_test");
    rtst_log("Testing padmat clone operations.\n");

    for (i=0; i<rtst_numtests; i++)
    {

	align = 1 << (lrand48() % 6);
	rows = (lrand48() % 100) + 1;
	cols = (lrand48() % 100) + 1;
	margin = (lrand48() % 100) + 1;
	lda = cols + margin;

	len = cols * rows;
        orig_p = malloc(len);
        rtst_urand_i8i8_vi8(len, -0x80, 0x7f, (rtst_int8 *) orig_p);
/*
	rtst_log("Cloning padmat, rows=%lu cols=%lu lda=%lu align=%d\n",
		     (unsigned long) rows, (unsigned long) cols,
		     (unsigned long) lda, align);
*/
	copy_p = rtst_padmat_clone(rows, cols, cols, lda, align, orig_p);

	orow_p = orig_p;
	crow_p = copy_p;
	for (j=0; j<rows; j++)
	{
	    rtst_checkeq_vi8vi8(cols, orow_p, crow_p);
	    orow_p += cols;
	    crow_p += lda;
	}

        free(orig_p);
        rtst_padmat_del(copy_p);
    }
    rtst_passed();
}

void
checkeq_vf_test(void)
{
    /* Note: 0.0/0.0 gives a NaN */

    const float NaN_f = rtst_ieee_divide_ff_f(0.0f, 0.0f);
    const double NaN_d = rtst_ieee_divide_dd_d(0.0, 0.0);

    float f1[] = { 1.0, -1.0, 0.0 };
    float f2[] = { 1.0, -1.0, 1.0 };
    double d1[] = { 1.0, -1.0, 0.0 };
    double d2[] = { 1.0, -1.0, 1.0 };

    /* Check that NaNs are considered the same as each other */
    f1[2] = NaN_f;
    f2[2] = NaN_f;
    d1[2] = NaN_d;
    d2[2] = NaN_d;

    /*  Debugging use only.
    fprintf(stderr, "%g\n", (double) NaN_f);
    fprintf(stderr, "%g\n", NaN_d);
    */

    rtst_start("checkeq_vfvf_test");
    rtst_checkeq_vfvf(3, f1, f2);
    rtst_passed();

    rtst_start("checkeq_vdvd_test");
    rtst_checkeq_vdvd(3, d1, d2);
    rtst_passed();
}

void
checkeq_mf_test(void)
{
    /* Note: 0.0/0.0 gives a NaN */


    float f1[] = 
    {
	1.0, -1.0, 0.0,
	2.0, 3.0, 1.0
    };
    float f2[] =
    {
	1.0, -1.0, 0.1, 0.2,
	2.0, 3.0, 1.0, 1.1
    };
    double d1[] = 
    {
	1.0, -1.0, 0.0,
	2.0, 3.0, 1.0
    };
    double d2[] =
    {
	1.0, -1.0, 0.1, 0.2,
	2.0, 3.0, 1.0, 1.1
    };

    rtst_start("checkeq_mfmf_test");
    rtst_checkeq_mfmf(2, 2, 3, 4, f1, f2);
    rtst_passed();

    rtst_start("checkeq_mdmd_test");
    rtst_checkeq_mdmd(2, 2, 3, 4, d1, d2);
    rtst_passed();
}

void
checknear_test(void)
{
    enum
    {
	MAX_LEN = 64
    };
    const double MAX_TOL = 1.0;
    const double MAX_VAL = 10.0;
    size_t i, j;
    float af[MAX_LEN];
    float bf[MAX_LEN];
    double ad[MAX_LEN];
    double bd[MAX_LEN];

    rtst_start("checknear_test");
    for (i=0; i<rtst_numtests; i++)
    {
	size_t len = (unsigned long) lrand48() % MAX_LEN;

	rtst_urand_ff_vf(len, -MAX_VAL, MAX_VAL, af);
	rtst_urand_ff_vf(len, -MAX_TOL, MAX_TOL, bf);
	for (j=0; j<len; j++)
	{
	    bf[j] += af[j];
	}
	rtst_urand_dd_vd(len, -MAX_VAL, MAX_VAL, ad);
	rtst_urand_dd_vd(len, -MAX_TOL, MAX_TOL, bd);
	for (j=0; j<len; j++)
	{
	    bd[j] += ad[j];
	}
	rtst_checknear_fvfvf(len, MAX_TOL, af, bf);
	rtst_checknear_dvdvd(len, MAX_TOL, ad, bd);
    }
    rtst_passed();
}

void
checkrange_test(void)
{
    enum
    {
	MAX_LEN = 64
    };
    size_t i;

    rtst_start("checkrange_test");
    for (i=0; i<rtst_numtests; i++)
    {
	size_t len = (unsigned long) lrand48() % MAX_LEN;

	{
	    rtst_int16 a16[MAX_LEN];
	    rtst_int16 min16, max16;

	    min16 = (float) rtst_urand_i16i16_i16(-0x8000, 0x7fff);
	    max16 = (float) rtst_urand_i16i16_i16(-0x8000, 0x7fff);
	    if (min16>max16)
	    {
		rtst_int16 tmp;
		
		tmp = min16; min16 = max16; max16 = tmp;
	    }
	    rtst_urand_i16i16_vi16(len, min16, max16, a16);
	    rtst_checkrange_i16i16vi16(len, min16, max16, a16);
	}

	{
	    rtst_int32 a32[MAX_LEN];
	    rtst_int32 min32, max32;

	    min32 = (float) rtst_urand_i32i32_i32(-0x7fffffff-1, 0x7fffffff);
	    max32 = (float) rtst_urand_i32i32_i32(-0x7fffffff-1, 0x7fffffff);
	    if (min32>max32)
	    {
		rtst_int32 tmp;
		
		tmp = min32; min32 = max32; max32 = tmp;
	    }
	    rtst_urand_i32i32_vi32(len, min32, max32, a32);
	    rtst_checkrange_i32i32vi32(len, min32, max32, a32);
	}

	{
	    float af[MAX_LEN];
	    float minf, maxf;

	    minf = (float) rtst_urand_dd_d(-12345.6, 23456.7);
	    maxf = (float) rtst_urand_dd_d(0, 12334.5) + minf;
	    rtst_urand_ff_vf(len, minf, maxf, af);
	    rtst_checkrange_ffvf(len, minf, maxf, af);
	}

	{
	    double ad[MAX_LEN];
	    double mind, maxd;

	    mind = rtst_urand_dd_d(-12345.6, 23456.7);
	    maxd = rtst_urand_dd_d(0, 12334.5) + mind;
	    rtst_urand_dd_vd(len, mind, maxd, ad);
	    rtst_checkrange_ddvd(len, mind, maxd, ad);
	}

    }
    rtst_passed();
}

int
main(int argc, char* argv[])
{
    int arg;

    arg = rtst_args(argc, argv);
    rtst_start("realtest_test");
    rtst_log("numtests=%i arg=%i\n", rtst_numtests, arg);
    seed48(rtst_state48);

    checkeq_vf_test();
    checkeq_mf_test();
    checknear_test();
    checkrange_test();
    padvec_test();
    padvec_clone_test();
    padmat_test();
    padmat_clone_test();

    rtst_exit();
    /* rtst_exit should call exit with zero value, this catches */
    /* error if it doesn't. */
    return 1;
}
